#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['ninja_schema',
 'ninja_schema.orm',
 'ninja_schema.orm.utils',
 'ninja_schema.pydanticutils']

package_data = \
{'': ['*']}

install_requires = \
['Django >=3.1', 'pydantic', 'pydantic[email]']

extras_require = \
{'test': ['pytest',
          'pytest-cov',
          'pytest-django',
          'pytest-asyncio',
          'black',
          'isort',
          'flake8',
          'mypy',
          'django-stubs']}

setup(name='ninja-schema',
      version='0.12.2',
      description='Django Schema - Builds Pydantic Schemas from Django Models',
      author='Ezeudoh Tochukwu',
      author_email='tochukwu.ezeudoh@gmail.com',
      url='https://github.com/eadwinCode/ninja-schema',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )
